
#pragma once

#include <TNL/Meshes/Grid.h>
#include <TNL/String.h>


struct VoidOperation {
   public:

      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {}
};


struct GetEntityIsBoundaryOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         bool isBoundary = entity.isBoundary();

         if (isBoundary)
            isBoundary = false;
      }
};

struct GetEntityCoordinateOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         typename Entity::GridType::Coordinate coordinate = entity.getCoordinates();

         coordinate.x() += 1;
      }
};

struct GetEntityIndexOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         int index = entity.getIndex();

         index += 1;
      }
};

struct GetEntityBasisOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         typename Entity::GridType::Coordinate coordinate = entity.getBasis();

         coordinate.x() += 1;
      }
};

struct RefreshEntityOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         entity.refresh();
      }
};

struct GetMeshDimensionOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         typename Entity::GridType::Coordinate coordinate = entity.getMesh().getDimensions();

         coordinate.x() += 1;
      }
};

struct GetOriginOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         typename Entity::GridType::Point coordinate = entity.getMesh().getOrigin();

         coordinate.x() += 1;
      }
};

struct GetEntitiesCountsOperation {
   public:
      template<typename Entity>
      __cuda_callable__ inline
      static void exec(Entity& entity) {
         typename Entity::GridType::EntitiesCounts coordinate = entity.getMesh().getEntitiesCounts();

         coordinate.x() += 1;
      }
};
